from .elements import *

matchMax = match_max
matchAtMost = atMost = at_most = match_at_most
matchNum = matchAmt = match_amt = amt = num = match_num
matchRange = match_range
matchMoreThan = match_greater_than = matchGreaterThan = moreThan = more_than = match_more_than
matchAtLeast = match_min = matchMin = atLeast = at_least = match_at_least
lineStartsWith = line_start = lineStart = line_starts_with
stringStartsWith = string_start = stringStart = string_starts_with
lineEndsWith = line_end = lineEnd = line_ends_with
stringEndsWith = string_end = stringEnd = string_ends_with
stuff      = chunk
whiteChunk = whitechunk
anychar    = anything
anyChar    = anything
char       = anything
alpha      = letter
alphanum   = alpha_num = alpha_num
white      = whitechunk

anyAmt = any_amt = zeroOrMore = zero_or_more = at_least_none
anyBetween = any_between
wordChar = word_char
hexDigit = hex
octDigit = oct_digit
newline = newLine = new_line
spaceOrTab = space_or_tab
carriageReturn = carriage_return
verticalTab = vertical_tab
formFeed = form_feed
dot = period
intOrFloat = int_or_float
notWhitespace = not_whitespace
notDigit = not_digit
notWord = not_word
anyof = any_of = anyOf = oneOf = one_of = any_of
anyExcept = any_except
anyCharExcept = any_char_except
printableAndSpace = printable_and_space
ifFollowedBy = if_followed_by = if_proceded_by
ifExists = if_exists
ifNotFollowedBy = if_not_followed_by = if_not_proceded_by
ifPrecededBy = if_preceded_by
ifNotPrecededBy = if_not_preceded_by
ifEnclosedWith = if_enclosed_by = ifEnclosedBy = if_enclosed_with
ifProcededBy = if_proceded_by
ifNotProcededBy = if_not_proceded_by
passiveGroup = passive_group
sameAs = same_as = earlierGroup = sameAsGroup = same_as_group = earlier_group
exactly = isExactly = is_exactly
oneOrNone = one_or_none = opt = optional
oneOrMore = one_or_more = atLeastOne = atLeast1 = at_least_1 = at_least_one
noneOrMore = none_or_more = atLeastNone = at_least_0 = atLeast0 = at_least_none
ascii = a = ASCII
dotall = s = DOTALL
ignorecase = i = ignoreCase = ignore_case = IGNORECASE
locale = L = LOCALE
multiline = m = MULTILINE
# Useful combinations
integer = signed
literallyAnything = literally_anything
wordBoundary = word_boundary
notWordBoundary = not_word_boundary

replaceGroup = replace_group = rgroup
replaceAll = replace_all = replaceEntire = replace_entire
